#include "docnomanager.h"
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <assert.h>
#include <db.h>

#define DB_ERR(err) if (err) {cerr << "docnomanager: db err " << db_strerror(err) << endl; exit(2);}

DocNoManager *DocNoManager::_instance = 0;

DocNoManager*
DocNoManager::Instance()
{
  return _instance;
}

DocNoManager::DocNoManager(string no2name, string name2no, bool update)
{
  _no2name = no2name;
  _name2no = name2no;
  _maxfo = 0;
  _instance = this;
  _update = update;
  _no2name_db = NULL;
  init ();
}

void
DocNoManager::init()
{
  //set up _no2name_db
  int err =  db_create(&_no2name_db, NULL, 0);
  DB_ERR(err);

  err = db_create(&_name2no_db, NULL, 0);
  DB_ERR(err);

  //get the max file number
  if (!_update) {
    err = _no2name_db->set_bt_compare(_no2name_db, &DocNoManager::compare_DID);
    DB_ERR (err);	
    err = _no2name_db->open(_no2name_db, NULL, _no2name.c_str(), NULL,
	DB_BTREE,DB_TRUNCATE|DB_CREATE, 0666);
    DB_ERR (err);
  
    err = _name2no_db->open(_name2no_db, NULL, _name2no.c_str(), NULL,
	DB_HASH, DB_TRUNCATE|DB_CREATE, 0666);
    DB_ERR(err);

    _maxfo = 0;
  } else {
    _no2name_db->set_bt_compare (_no2name_db, &DocNoManager::compare_DID);
    DB_ERR (err);
    err = _no2name_db->open(_no2name_db, NULL, _no2name.c_str(), NULL, 
	DB_BTREE, DB_CREATE, 0666);
    DB_ERR (err);

    err = _name2no_db->open(_name2no_db, NULL, _name2no.c_str(), NULL,
	        DB_HASH, DB_CREATE, 0666);
    DB_ERR(err);

    DBC *cursor;
    DBT key, data;
    bzero(&key, sizeof(key));
    bzero(&data, sizeof(data));
    err = _no2name_db->cursor(_no2name_db, NULL, &cursor, 0);
    DB_ERR(err);
    err = cursor->c_get(cursor, &key, &data, DB_LAST);
    if (err) {
      cerr << "docnomanager: WARNING " << db_strerror(err) << endl;
      _maxfo = 0;
    } else 
      memcpy (&_maxfo, key.data, sizeof(_maxfo));
  }
}

DocNoManager::~DocNoManager()
{
  cleanup();
}

void
DocNoManager::cleanup()
{
  if (_no2name_db) {
    DBC *cursor;
    DBT key, data;
    bzero(&key, sizeof(key));
    bzero(&data, sizeof(data));
    int err = _no2name_db->cursor(_no2name_db, NULL, &cursor, 0);
    DB_ERR(err);
    err = cursor->c_get(cursor, &key, &data, DB_LAST);
    assert (!err);
    if (_maxfo != (*(DID *)key.data))
      cerr << "mohahah _maxfo " << _maxfo << " db " << *(DID *)key.data << endl;
    _no2name_db->close(_no2name_db,0);
    _no2name_db = NULL;
  }

  if (_name2no_db) {
    _name2no_db->close(_name2no_db,0);
    _name2no_db = NULL;
  }
}

DID
DocNoManager::getDocNo(string fname)
{
   DBT key, data;

   bzero(&key,sizeof(key));
   bzero(&data,sizeof(data));


   key.data = (void *)fname.c_str();
   key.size = strlen(fname.c_str())+1;

   _maxfo++; 
    data.data = (void *)&_maxfo;
    data.size = sizeof(_maxfo);
  
    int err = _name2no_db->put(_name2no_db, NULL, &key, &data, DB_NOOVERWRITE);
    if (err) {
       //duplicate document names!
       //fprintf(stderr,"duplicate doc name %s prev no %u\n", fname.c_str (), (*(DID *)data.data));
       _maxfo--;
       return 0;
  
    } else {

       key.data = &_maxfo;
       key.size = sizeof(_maxfo);

       data.data = (void *)fname.c_str();
       data.size = strlen(fname.c_str())+1;
     
       err = _no2name_db->put(_no2name_db, NULL, &key,&data,DB_NOOVERWRITE);

       if (err) {
	  cerr << "donomanager: " << db_strerror(err)
	     << " " << _maxfo << ":" << fname << endl;
	  return 0;
       } 
    }
    return  _maxfo;
}
