/*
 * input: a stream of <char *, ind_offset> pairs
 * Create the DB file mapping words to offset into the inverted file
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "mkdb.h"
#include "args.h"

struct timeval tv0;

void printrusage();
int zgets(char s[], int max, FILE *fp);

int
main(int argc, char *argv[])
{
  DB *db;
  DBT key, data;
  unsigned long long distinctwords = 0;

  gettimeofday(&tv0, (struct timezone *) 0);

  char *config = "mkdb.config";

  for (int i = 1; i < argc; i++) {
    if (strcmp(argv[i],"-f")==0 && i+1 < argc) {
      i++;
      config = argv[i];
    }else{
      fprintf(stderr,"usage: mkdb [-f configfile]\n");
      exit(1);
    }
  }

  char dbfile[100];
  Args *a = new Args(config);
  strncpy(dbfile,a->sget("prefix","ind").c_str(),90);
  strcat(dbfile,"-w2p.db");

  unsigned maxwordlen = a->nget<unsigned>("maxwordlen",100);
  char word[maxwordlen+1];

  int err = db_create(&db, NULL, 0);
  if (err) {
    fprintf(stderr,"failed to create db\n");
    exit(2);
  }
  err = db->open(db, NULL, dbfile, NULL,
      DB_BTREE, DB_TRUNCATE|DB_CREATE, 0666);
  if (err) {
    fprintf(stderr,"failed to open db %s\n", dbfile);
    exit(1);
  }

  ind_offset index = 0;
  bzero(&key,sizeof(key));
  bzero(&data,sizeof(data));
  while(zgets(word, sizeof(word), stdin) > 0){
    fread(&index, sizeof(index), 1, stdin);
    key.data = word;
    key.size = strlen(word) + 1;
    data.data = &index;
    data.size = sizeof(index);
    if(db->put(db, NULL, &key, &data, DB_NOOVERWRITE) != 0){
      fprintf(stderr, "mkdb: db->put failed\n");
      exit(1);
    }
    distinctwords++;
  }
  
  if(db->close(db,0) != 0){
    fprintf(stderr, "mkdb: db close failed\n");
    exit(1);
  }

  fprintf(stderr, "mkdb: %qu distinct\n",
          (unsigned long long) distinctwords);
  printrusage();

  delete a;
  exit(0);
}

/*
 * Read a null-terminated string.
 * Returns EOF on error, otherwise n chars read.
 */
int
zgets(char s[], int max, FILE *fp)
{
  int i = 0, c;

  while((c = getc(fp)) != EOF){
    if(i >= max){
      fprintf(stderr, "mkdb: intermediate word too long\n");
      exit(1);
    }
    s[i++] = c;
    if(c == '\0')
      break;
  }
  if(i == 0 && c == EOF)
    return(EOF);
  return(i);
}

void
printrusage()
{
  struct rusage ru;
  struct timeval tv;

  gettimeofday(&tv, (struct timezone *) 0);
  if(getrusage(RUSAGE_SELF, &ru) < 0){
    perror("nisort: getrusage");
    return;
  }
  fprintf(stderr, "mkdb: %.2fu %.2fs %.2fr %ldi %ldo\n",
          ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec / 1000000.0),
          ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec / 1000000.0),
          (tv.tv_sec + tv.tv_usec/1000000.0) -
            (tv0.tv_sec + tv0.tv_usec/1000000.0),
          ru.ru_inblock,
          ru.ru_oublock);
}
